import lldb

import ctypes

import math



def BaseObject(valueObject, dictionary):

    

    target = valueObject.GetTarget()

    Vector_Type = target.FindFirstType('Vector')

    

    pos = valueObject.CreateChildAtOffset("pos", 1152, Vector_Type)

    rot = valueObject.CreateChildAtOffset("rot", 1128, Vector_Type)

    scale = valueObject.CreateChildAtOffset("scale", 1176, Vector_Type)

    

    return BaseList2D(valueObject, dictionary) + " , " + Vector(pos, dictionary) + " , " + Vector(rot, dictionary) + " , " + Vector(scale, dictionary)



def BaseDocument(valueObject, dictionary):



    filestr = valueObject.GetChildMemberWithName('dummy1')

    return String(filestr, dictionary)



def Matrix(valueObject, dictionary):

    

    off = valueObject.GetChildMemberWithName('off')

    

    error = lldb.SBError()

    

    off_x = off.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

    off_y = off.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

    off_z = off.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

    v1 = valueObject.GetChildMemberWithName('v1')

    

    v1_x = v1.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

    v1_y = v1.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

    v1_z = v1.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

    v2 = valueObject.GetChildMemberWithName('v2')

    

    v2_x = v2.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

    v2_y = v2.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

    v2_z = v2.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

    v3 = valueObject.GetChildMemberWithName('v3')

    

    v3_x = v3.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

    v3_y = v3.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

    v3_z = v3.GetChildMemberWithName('z').GetData().GetDouble(error, 0)



    return '[ (' + str(off_x) + ' , ' + str(off_y) + ' , ' + str(off_z) + ') , (' + str(v1_x) + ' , ' + str(v1_y) + ' , ' + str(v1_z) + ') , (' + str(v2_x) + ' , ' + str(v2_y) + ' , ' + str(v2_z) + ') , (' + str(v3_x) + ' , ' + str(v3_y) + ' , ' + str(v3_z) + ') ' + ']'



def Vector(valueObject, dictionary):

    

    error = lldb.SBError()



    x = valueObject.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

    y = valueObject.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

    z = valueObject.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

    len = math.sqrt(x * x + y * y + z * z)

    

    return '(' + str(x) + ' , ' + str(y) + ' , ' + str(z) + ') , ' + str(len)



def FileName(valueObject, dictionary):



    return String(valueObject, dictionary)



def BaseShader(valueObject, dictionary):



    return BaseList2D(valueObject, dictionary)



def BasePlugin(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def BaseSceneHook(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def BaseSceneLoader(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def BaseSceneSaver(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def BaseTag(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def BaseVideoPost(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def CCurve(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def CTrack(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)



def GvNode(valueObject, dictionary):

    

    return BaseList2D(valueObject, dictionary)





# c4d String

def String(valueObject, dictionary):

  

    error = lldb.SBError()



    sbtype = valueObject.GetType()



    voidPtrType = sbtype.GetBasicType(lldb.eBasicTypeVoid).GetPointerType()

    int32Type = sbtype.GetBasicType(lldb.eBasicTypeInt)

    uint32Type = sbtype.GetBasicType(lldb.eBasicTypeUnsignedInt)

    intType = sbtype.GetBasicType(lldb.eBasicTypeLong)

    uniCharPtrType = sbtype.GetBasicType(lldb.eBasicTypeChar16).GetPointerType()



    referenceCnt = 0

    weakReferences = 0

    ptr = valueObject.CreateChildAtOffset("ptr", 8, uniCharPtrType)



    if ptr.GetValueAsUnsigned(0) == 0:

        #Do not call CreateChildAtOffset() for offset 0 before we have determined whether this might be a referenced

        #string or not. Once a child has been created we can't create a new one for the same offset with a different type.



        # TODO: (sven) check this with seb

        #For some reason strRef is just an int and therefore objPtr is incorrect when I'm using the following code

        #strRef = valueObject.CreateChildAtOffset("strRef", 0, voidPtrType)

        #objPtr = strRef.GetValueAsUnsigned(0)

        

        strRef = valueObject.CreateChildAtOffset("strRef", 0, voidPtrType).AddressOf()

        capacity = 0



        objPtr = valueObject.CreateValueFromAddress("objPtr", strRef.GetValueAsUnsigned(), voidPtrType).GetValueAsUnsigned()

        if objPtr == 0:

            cnt = 0

            ptr = 0

        else:

            referenceCnt = valueObject.CreateValueFromAddress("referenceCnt", objPtr, uint32Type).GetValueAsSigned()

            objPtr = objPtr + 4

            weakReferences = valueObject.CreateValueFromAddress("weakReferences", objPtr, uint32Type).GetValueAsUnsigned()

            objPtr = objPtr + 4

            ptr = valueObject.CreateValueFromAddress("ptr", objPtr, uniCharPtrType)

            objPtr = objPtr + 8

            cnt = valueObject.CreateValueFromAddress("cnt", objPtr, intType).GetValueAsSigned()

            objPtr = objPtr + 8

            capacity = valueObject.CreateValueFromAddress("capacity", objPtr, intType).GetValueAsSigned()

    else:

        capacity = valueObject.CreateChildAtOffset("capacity", 0, int32Type).GetData().GetSignedInt32(error, 0)

        cnt = valueObject.CreateChildAtOffset("cnt", 4, int32Type).GetData().GetSignedInt32(error, 0)



    #Check for garbage values (this might be called on an uninitialized String).

    #If capacity is a very large value something might be wrong (and it must not be smaller than cnt).

    if capacity > 1073741824:

        cnt = 0



    if cnt > capacity:

        cnt = 0



    #Limit number of characters (this might be called on an uninitialized String).

    displayCnt = cnt

    if cnt > 128:

        displayCnt = 128



    s = u'"'

    for i in xrange(0, displayCnt):

        data_val = ptr.GetPointeeData(i, 1)

        newchar = data_val.GetUnsignedInt16(error, 0)

        size = data_val.GetByteSize()

        s += unichr(newchar)

    s += u'"'



    if cnt > 0 and ptr != 0:

        s += " (%i @ %s)" % (cnt, hex(ptr.GetValueAsUnsigned()))



    return s.encode('utf-8')



def Time(valueObject, dictionary):



    error = lldb.SBError()



    num = valueObject.GetChildMemberWithName('numerator').GetData().GetDouble(error, 0)

    den = valueObject.GetChildMemberWithName('denominator').GetData().GetDouble(error, 0)



    sec = num / den



    return str(num) + '/' + str(den) + ' ' + str(sec) + ' s'



def Polygon(valueObject, dictionary):



    error = lldb.SBError()



    a = valueObject.GetChildMemberWithName('a').GetData().GetUnsignedInt32(error, 0)

    b = valueObject.GetChildMemberWithName('b').GetData().GetUnsignedInt32(error, 0)

    c = valueObject.GetChildMemberWithName('c').GetData().GetUnsignedInt32(error, 0)

    d = valueObject.GetChildMemberWithName('d').GetData().GetUnsignedInt32(error, 0)

    

    if c == d:

    

        id = "Tri"

    

    if c != d:

            

        id = "Quad"



    return str(a) + ' , ' + str(b) + ' , ' + str(c) + ' , ' + str(d) + ' , ' + str(id)



def MinMax(valueObject, dictionary):



    error = lldb.SBError()

    

    used = valueObject.GetChildMemberWithName('used').GetData().GetUnsignedInt32(error, 0)

    

    if used == 1:

        

        min = valueObject.GetChildMemberWithName('min')

        minx = min.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

        miny = min.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

        minz = min.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

        max = valueObject.GetChildMemberWithName('max')

        maxx = max.GetChildMemberWithName('x').GetData().GetDouble(error, 0)

        maxy = max.GetChildMemberWithName('y').GetData().GetDouble(error, 0)

        maxz = max.GetChildMemberWithName('z').GetData().GetDouble(error, 0)

    

        return '( (' + str(minx) + ' , ' + str(miny) + ' , ' + str(minz) + ') , (' + str(maxx) + ' , ' + str(maxy) + ' , ' + str(maxz) + ') )'

    

    else:

    

        return "(undefined)"



def DateTime(valueObject, dictionary):



    error = lldb.SBError()



    year = valueObject.GetChildMemberWithName('year').GetData().GetUnsignedInt32(error, 0)

    month = valueObject.GetChildMemberWithName('month').GetData().GetUnsignedInt32(error, 0)

    day = valueObject.GetChildMemberWithName('day').GetData().GetUnsignedInt32(error, 0)

    hour = valueObject.GetChildMemberWithName('hour').GetData().GetUnsignedInt32(error, 0)

    minute = valueObject.GetChildMemberWithName('minute').GetData().GetUnsignedInt32(error, 0)

    second = valueObject.GetChildMemberWithName('second').GetData().GetUnsignedInt32(error, 0)



    return str(year) + ' , ' + str(month) + ' , ' + str(day) + ' , ' + str(hour) + ' , ' + str(minute) + ' , ' + str(second)



def DescLevel(valueObject, dictionary):



    error = lldb.SBError()



    id = valueObject.GetChildMemberWithName('id').GetData().GetUnsignedInt32(error, 0)

    dtype = valueObject.GetChildMemberWithName('dtype').GetData().GetUnsignedInt32(error, 0)

    creator = valueObject.GetChildMemberWithName('creator').GetData().GetUnsignedInt32(error, 0)



    return str(id) + ' , ' + str(dtype) + ' , ' + str(creator)



def BaseList2D(valueObject, dictionary):



    error = lldb.SBError()

    

    target = valueObject.GetTarget()

    strType = target.FindFirstType('String')

    

    char_type = strType.GetBasicType(lldb.eBasicTypeChar)

    charptr_type = char_type.GetPointerType()

    int_type = charptr_type.GetBasicType(lldb.eBasicTypeInt)



    name = valueObject.CreateChildAtOffset("name", 576, strType)

    plugindata = valueObject.CreateChildAtOffset("plugindata", 72, charptr_type)

    id = plugindata.CreateChildAtOffset("id", 8, int_type).GetData().GetUnsignedInt32(error, 0)

    

    return String(name, dictionary) + " " + str(id)



def IconData(valueObject, dictionary):



    error = lldb.SBError()

    

    x = valueObject.GetChildMemberWithName('x').GetData().GetUnsignedInt32(error, 0)

    y = valueObject.GetChildMemberWithName('y').GetData().GetUnsignedInt32(error, 0)

    w = valueObject.GetChildMemberWithName('w').GetData().GetUnsignedInt32(error, 0)

    h = valueObject.GetChildMemberWithName('h').GetData().GetUnsignedInt32(error, 0)

    flags = valueObject.GetChildMemberWithName('flags').GetData().GetUnsignedInt32(error, 0)

    

    return str(x) + ' , ' + str(y) + ' , ' + str(w) + ' , ' + str(h)



def BaseBitmap(valueObject, dictionary):

    

    error = lldb.SBError()

    

    target = valueObject.GetTarget()

    strType = target.FindFirstType('String')

    

    char_type = strType.GetBasicType(lldb.eBasicTypeChar)

    charptr_type = char_type.GetPointerType()

    int_type = charptr_type.GetBasicType(lldb.eBasicTypeInt)

    

    Width = valueObject.CreateChildAtOffset("Width", 152, int_type).GetData().GetUnsignedInt32(error, 0)

    

    Height = valueObject.CreateChildAtOffset("Height", 156, int_type).GetData().GetUnsignedInt32(error, 0)

    

    return str(Width) + " , " + str(Height)





def __lldb_init_module(debugger, dict):

    

    debugger.HandleCommand('type summary add BaseObject -F c4dapi.BaseObject')

    debugger.HandleCommand('type summary add BaseDocument -F c4dapi.BaseDocument')

    debugger.HandleCommand('type summary add LVector -F c4dapi.Vector')

    debugger.HandleCommand('type summary add String -F c4dapi.String')

    debugger.HandleCommand('type summary add BaseTime -F c4dapi.Time')

    debugger.HandleCommand('type summary add CPolygon -F c4dapi.Polygon')

    debugger.HandleCommand('type summary add SVector -F c4dapi.Vector')

    debugger.HandleCommand('type summary add Vector64 -F c4dapi.Vector')

    debugger.HandleCommand('type summary add Vector32 -F c4dapi.Vector')

    debugger.HandleCommand('type summary add Matrix64 -F c4dapi.Matrix')

    debugger.HandleCommand('type summary add LMatrix -F c4dapi.Matrix')

    debugger.HandleCommand('type summary add Matrix32 -F c4dapi.Matrix')

    debugger.HandleCommand('type summary add SMatrix -F c4dapi.Matrix')

    debugger.HandleCommand('type summary add SMinMax -F c4dapi.MinMax')

    debugger.HandleCommand('type summary add LMinMax -F c4dapi.MinMax')

    debugger.HandleCommand('type summary add DateTime -F c4dapi.DateTime')

    debugger.HandleCommand('type summary add DescLevel -F c4dapi.DescLevel')

    debugger.HandleCommand('type summary add BaseList2D -F c4dapi.BaseList2D')

    debugger.HandleCommand('type summary add BaseShader -F c4dapi.BaseShader')

    debugger.HandleCommand('type summary add BasePlugin -F c4dapi.BasePlugin')

    debugger.HandleCommand('type summary add BaseSceneHook -F c4dapi.BaseSceneHook')

    debugger.HandleCommand('type summary add BaseSceneLoader -F c4dapi.BaseSceneLoader')

    debugger.HandleCommand('type summary add BaseSceneSaver -F c4dapi.BaseSceneSaver')

    debugger.HandleCommand('type summary add BaseTag -F c4dapi.BaseTag')

    debugger.HandleCommand('type summary add BaseVideoPost -F c4dapi.BaseVideoPost')

    debugger.HandleCommand('type summary add CCurve -F c4dapi.CCurve')

    debugger.HandleCommand('type summary add CTrack -F c4dapi.CTrack')

    debugger.HandleCommand('type summary add GvNode -F c4dapi.GvNode')

    debugger.HandleCommand('type summary add Filename -F c4dapi.FileName')

    debugger.HandleCommand('type summary add IconData -F c4dapi.IconData')

    debugger.HandleCommand('type summary add BaseBitmap -F c4dapi.BaseBitmap')