/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

//----------------------------------------------------
//Falloff controller system
#ifndef C4D_FALLOFFDATA_H__
#define C4D_FALLOFFDATA_H__

#ifdef __API_INTERN__
abc def xyz
#endif

#include "ge_math.h"
#include "ofalloff_standard.h"
#include "c4d_basedata.h"
#include "c4d_basecontainer.h"

class HandleInfo;

#define FALLOFFCAGE		 Vector(1.0f, 1.0f, 0.0f)
#define FALLOFFFALLOFF Vector(0.7f, 0.1f, 0.0f)

#define	ID_C4DFALLDATALIB	1019551
#define ID_C4DFALLOFFLIB	1019541

struct FalloffDataData
{
public:
	FalloffDataData(void)
	{
		_mat	 = Matrix();
		_imat = ~_mat;
		_falloff = _scale = _strength = 1.0;
		_size = Vector(1.0);
		_invert	= _clamped = _useSpline = false;
		_visible = true;

		_axis	= 0;
		_slice = _torusRadius = _radius = _inverseRadius = _inverseFalloff = 1.0;
		_function	= FALLOFF_FUNCTION_SPLINE;
		_spherical = false;

		_userData = nullptr;
	}
	~FalloffDataData(void) { }

	Bool operator ==(const FalloffDataData& b)
	{
		return _mat == b._mat && _imat == b._imat && _falloff == b._falloff && _scale == b._scale && _strength == b._strength
					 && _size == b._size && _invert == b._invert && _clamped == b._clamped && _visible == b._visible && _useSpline == b._useSpline
					 && _nodemat == b._nodemat && _inverseNodeMat == b._inverseNodeMat && _offset == b._offset && _axis == b._axis && _slice == b._slice
					 && _torusRadius == b._torusRadius && _function == b._function && _spherical == b._spherical && _radius == b._radius
					 && _inverseRadius == b._inverseRadius && _inverseFalloff == b._inverseFalloff && _userData == b._userData;
	}

	Bool operator !=(const FalloffDataData& b)
	{
		return !(*this == b);
	}

	Matrix _mat, _imat;
	Float	 _falloff, _scale, _strength;
	Vector _size;
	Bool	 _invert, _clamped, _visible, _useSpline;
	Matrix _nodemat, _inverseNodeMat;
	Vector _offset;
	Int32	 _axis;
	Float	 _slice;
	Float	 _torusRadius;
	Int32	 _function;
	Bool	 _spherical;
	Float	 _radius, _inverseRadius;
	Float	 _inverseFalloff;
	void*	 _userData;
};

class FalloffData : public BaseData
{
public:
	virtual Bool Init(FalloffDataData& falldata, BaseContainer* bc);
	virtual Bool InitFalloff(BaseContainer* bc, FalloffDataData& falldata);
	virtual void Sample(const Vector& p, const FalloffDataData& data, Float* res);
	virtual void FreeFalloff(FalloffDataData& falldata);
	virtual Bool CheckDirty(BaseContainer* bc);
	virtual Bool GetDVisible(DescID& id, BaseContainer* bc, BaseContainer* desc_bc);
	virtual Int32	GetHandleCount(BaseContainer* bc, const FalloffDataData& data);
	virtual void GetHandle(BaseContainer* bc, Int32 i, HandleInfo& info, const FalloffDataData& data);
	virtual void SetHandle(BaseContainer* bc, Int32 i, Vector p, const FalloffDataData& data);
	virtual Bool Draw(const FalloffDataData& data, DRAWPASS drawpass, BaseDraw* bd, BaseDrawHelp* bh);
	virtual Bool Message(Int32 type, BaseContainer* bc, void* m_data);
};

class C4D_Falloff
{
private:
	C4D_Falloff(void) { }
	~C4D_Falloff(void) { }

public:
	BaseContainer*					GetContainerInstance(void);

	void SetDirty(void);
	Int32	GetDirty(BaseContainer* bc = nullptr);
	void SetMg(Matrix mg);
	Matrix GetMg(void);
	void SetData(FalloffDataData falldata);
	FalloffDataData	GetData(void);
	Bool SetMode(Int32 type, BaseContainer* bc = nullptr);
	Int32	GetMode(void);
	void SetTime(BaseTime time, BaseContainer* bc = nullptr);

	Bool InitFalloff(BaseContainer* bc = nullptr, BaseDocument* doc = nullptr, BaseObject* op = nullptr);
	void Sample(const Vector& p, Float* result, Bool usespline = true, Float weight = 0.0);

	void MultiSample(const Vector* p, Float* result, Int32 count, Bool usespline = true, Float weight = 0.0);

	Bool AddFalloffToDescription(Description* description, BaseContainer* bc = nullptr);

	Bool Message(Int32 id, BaseContainer* bc = nullptr, void* m_data = nullptr);
	Int32	GetHandleCount(BaseContainer* bc = nullptr);
	void GetHandle(Int32 i, BaseContainer* bc, HandleInfo& info);
	void SetHandle(Int32 i, Vector p, BaseContainer* bc, const HandleInfo& info);
	Bool Draw(BaseDraw* bd, BaseDrawHelp* bh, DRAWPASS drawpass, BaseContainer* bc = nullptr);
	Bool CopyTo(C4D_Falloff* dest);

	//Autoallocation code
	static C4D_Falloff*			Alloc(void);
	static void	Free(C4D_Falloff*& node);
};

struct FALLOFFPLUGIN;

typedef FalloffData* FalloffDataAllocator (void);

void FillFalloffPlugin(FALLOFFPLUGIN* fp, Int32 info, FalloffDataAllocator* g);
Bool RegisterFalloffPlugin(Int32 id, const String& str, Int32 info, FalloffDataAllocator* g, const String& description);
void* AllocFalloff(Int32 type);

#endif // C4D_FALLOFFDATA_H__
