/////////////////////////////////////////////////////////////
// CINEMA 4D SDK                                           //
/////////////////////////////////////////////////////////////
// (c) MAXON Computer GmbH, all rights reserved            //
/////////////////////////////////////////////////////////////

#ifndef C4D_REFLECTION_H__
#define C4D_REFLECTION_H__

#include "ge_prepass.h"	// important
#include "c4d_string.h"

// Layer Flags
#define REFLECTION_FLAG_NONE													0x00
//
#define REFLECTION_FLAG_SELECTED											0x01
#define REFLECTION_FLAG_ACTIVE												0x02
#define REFLECTION_FLAG_TAB														0x04
#define REFLECTION_FLAG_TRANSPARENCY									0x08
//
#define REFLECTION_FLAG_SEPARATE											0x10
#define REFLECTION_FLAG_READY													0x20

// Additive Modes
#define REFLECTION_ADDITIVE_MODE_AVG									0
#define REFLECTION_ADDITIVE_MODE_MIN_EX								1
#define REFLECTION_ADDITIVE_MODE_MAX									2
#define REFLECTION_ADDITIVE_MODE_ADD									3
#define REFLECTION_ADDITIVE_MODE_METAL								4

// Distributions
#define REFLECTION_DISTRIBUTION_SIMPLE								0
#define REFLECTION_DISTRIBUTION_SPECULAR_BLINN				1
//
#define REFLECTION_DISTRIBUTION_BECKMANN							2
#define REFLECTION_DISTRIBUTION_GGX										3
#define REFLECTION_DISTRIBUTION_PHONG									4
#define REFLECTION_DISTRIBUTION_WARD									5
#define REFLECTION_DISTRIBUTION_ANISOTROPIC						6
//
#define REFLECTION_DISTRIBUTION_LAMBERTIAN						7
#define REFLECTION_DISTRIBUTION_ORENNAYAR							8
//
#define REFLECTION_DISTRIBUTION_IRAWAN								9
// ---
#define REFLECTION_DISTRIBUTION_SPECULAR_PHONG				10

// Bump Mode
#define REFLECTION_BUMP_MODE_DEFAULT									0
#define REFLECTION_BUMP_MODE_BUMP											1
#define REFLECTION_BUMP_MODE_NORMAL										2

// Projection
#define REFLECTION_PROJECTION_PLANAR									0
#define REFLECTION_PROJECTION_RADIAL									1
#define REFLECTION_PROJECTION_NONE										2

// Mirror
#define REFLECTION_MIRROR_NONE												0
#define REFLECTION_MIRROR_U														1
#define REFLECTION_MIRROR_V														2
#define REFLECTION_MIRROR_UV													3

// Patterns
#define REFLECTION_PATTERN_NONE												0
#define REFLECTION_PATTERN_CIRCULAR										1
#define REFLECTION_PATTERN_BOX												2
#define REFLECTION_PATTERN_DIAGONAL										3
#define REFLECTION_PATTERN_LATTICE										4
#define REFLECTION_PATTERN_CUSTOM											5

// Scratches
#define REFLECTION_SCRATCHES_NONE											0
#define REFLECTION_SCRATCHES_PRIMARY									1
#define REFLECTION_SCRATCHES_SECONDARY								2
#define REFLECTION_SCRATCHES_BOTH											3

// Cloth Presets
#define REFLECTION_CLOTH_PRESET_CUSTOM								0
#define REFLECTION_CLOTH_PRESET_A											1
#define REFLECTION_CLOTH_PRESET_B											2
#define REFLECTION_CLOTH_PRESET_C											3
#define REFLECTION_CLOTH_PRESET_D											4
#define REFLECTION_CLOTH_PRESET_E											5
#define REFLECTION_CLOTH_PRESET_F											6

// Cloth Patterns
#define REFLECTION_CLOTH_PATTERN_COTTON_DENIM					0
#define REFLECTION_CLOTH_PATTERN_COTTON_TWILL					1
#define REFLECTION_CLOTH_PATTERN_POLYESTER_LINING			2
#define REFLECTION_CLOTH_PATTERN_SILK_CHARMEUSE				3
#define REFLECTION_CLOTH_PATTERN_SILK_SHANTUNG				4
#define REFLECTION_CLOTH_PATTERN_WOOL_GABARDINE				5

// Cloth Quality
#define REFLECTION_CLOTH_QUALITY_LOW									0
#define REFLECTION_CLOTH_QUALITY_MEDIUM								1
#define REFLECTION_CLOTH_QUALITY_HIGH									2

// Fresnel
#define REFLECTION_FRESNEL_NONE												0
#define REFLECTION_FRESNEL_DIELECTRIC									1
#define REFLECTION_FRESNEL_CONDUCTOR									2

// Dielectrics
#define REFLECTION_FRESNEL_DIELECTRIC_CUSTOM					0
#define REFLECTION_FRESNEL_DIELECTRIC_ASPHALT					1
#define REFLECTION_FRESNEL_DIELECTRIC_BEER						2
#define REFLECTION_FRESNEL_DIELECTRIC_DIAMOND					3
#define REFLECTION_FRESNEL_DIELECTRIC_EMERALD					4
#define REFLECTION_FRESNEL_DIELECTRIC_ETHANOL					5
#define REFLECTION_FRESNEL_DIELECTRIC_GLASS						6
#define REFLECTION_FRESNEL_DIELECTRIC_JADE						7
#define REFLECTION_FRESNEL_DIELECTRIC_MILK						8
#define REFLECTION_FRESNEL_DIELECTRIC_OIL_VEGETABLE		9
#define REFLECTION_FRESNEL_DIELECTRIC_PEARL						10
#define REFLECTION_FRESNEL_DIELECTRIC_PET							11
#define REFLECTION_FRESNEL_DIELECTRIC_PLEXIGLASS			12
#define REFLECTION_FRESNEL_DIELECTRIC_RUBY						13
#define REFLECTION_FRESNEL_DIELECTRIC_SAPPHIRE				14
#define REFLECTION_FRESNEL_DIELECTRIC_WATER						15
#define REFLECTION_FRESNEL_DIELECTRIC_WATER_ICE				16
#define REFLECTION_FRESNEL_DIELECTRIC_WHISKEY					17

// Metals
#define REFLECTION_FRESNEL_METAL_CUSTOM								0
#define REFLECTION_FRESNEL_METAL_ALUMINUM							1
#define REFLECTION_FRESNEL_METAL_BERYLLIUM						2
#define REFLECTION_FRESNEL_METAL_CHROMIUM							3
#define REFLECTION_FRESNEL_METAL_COPPER								4
#define REFLECTION_FRESNEL_METAL_COPPER_OXIDE					5
#define REFLECTION_FRESNEL_METAL_GOLD									6
#define REFLECTION_FRESNEL_METAL_IRIDIUM							7
#define REFLECTION_FRESNEL_METAL_IRON									8
#define REFLECTION_FRESNEL_METAL_LITHIUM							9
#define REFLECTION_FRESNEL_METAL_MAGNESIUM_OXIDE			10
#define REFLECTION_FRESNEL_METAL_MERCURY							11
#define REFLECTION_FRESNEL_METAL_NICKEL								12
#define REFLECTION_FRESNEL_METAL_NIOBIUM							13
#define REFLECTION_FRESNEL_METAL_POTASSIUM						14
#define REFLECTION_FRESNEL_METAL_RHODIUM							15
#define REFLECTION_FRESNEL_METAL_SELENIUM							16
#define REFLECTION_FRESNEL_METAL_SILICON_CARBIDE			17
#define REFLECTION_FRESNEL_METAL_SILVER								18
#define REFLECTION_FRESNEL_METAL_SODIUM								19
#define REFLECTION_FRESNEL_METAL_TANTALUM							20
#define REFLECTION_FRESNEL_METAL_TELLURIUM						21
#define REFLECTION_FRESNEL_METAL_TIN_TELLURIDE				22
#define REFLECTION_FRESNEL_METAL_TITANIUM_NITRIDE			23
#define REFLECTION_FRESNEL_METAL_TUNGSTEN							24
#define REFLECTION_FRESNEL_METAL_VANADIUM							25

// Debug DU/DV
#define REFLECTION_DEBUG_DUDV_NORMAL									0
#define REFLECTION_DEBUG_DUDV_DEFAULT_U								1
#define REFLECTION_DEBUG_DUDV_DEFAULT_V								2
#define REFLECTION_DEBUG_DUDV_TANGENT_U								3
#define REFLECTION_DEBUG_DUDV_TANGENT_V								4
#define REFLECTION_DEBUG_DUDV_ANGLE										5

// Debug Scratch
#define REFLECTION_DEBUG_SCRATCH_NORMAL								0
#define REFLECTION_DEBUG_SCRATCH_DDU									1
#define REFLECTION_DEBUG_SCRATCH_DDV									2
#define REFLECTION_DEBUG_SCRATCH_BOTH									3
#define REFLECTION_DEBUG_SCRATCH_DELTA								4

// Debug Cloth
#define REFLECTION_DEBUG_CLOTH_NORMAL									0
#define REFLECTION_DEBUG_CLOTH_NOISE									1
#define REFLECTION_DEBUG_CLOTH_YARNS									2

// Maximum 16 layers as we need unique IDs for all of them
#define REFLECTION_MAX_LAYERS													16

// All Layers Data
#define REFLECTION_LAYER_DUMMY												0x0000
#define REFLECTION_LAYER_QUICKTAB											0x8000
#define REFLECTION_LAYER_ADD													0x8010
#define REFLECTION_LAYER_REMOVE												0x8020
#define REFLECTION_LAYER_COPY													0x8030
#define REFLECTION_LAYER_PASTE												0x8040
#define REFLECTION_LAYER_DUPLICATE										0x8050
#define REFLECTION_LAYER_DATA													0x80A0
#define REFLECTION_LAYER_GLOBAL_REFLECTION						0x80B0
#define REFLECTION_LAYER_GLOBAL_SPECULAR							0x80C0
#define REFLECTION_LAYER_GLOBAL_SEPARATE_PASS					0x80D0
#define REFLECTION_LAYER_GLOBAL_GROUP									0x80E0
// ---
#define REFLECTION_LAYER_IMPORTED											0x80F0 // Set to true if imported (unlikely)
#define REFLECTION_LAYER_ENABLED											0x80F1
#define REFLECTION_LAYER_PREVIEWS											0x80FF // Special ID (don't change)

// Layers data IDs and size
#define REFLECTION_LAYER_LAYER_DATA										0x80000 // Reserve the whole 0x80000 group
#define REFLECTION_LAYER_LAYER_SIZE										0x0200

// Layers (Maximum of 16 layers)
// IDs are added to the layer base data ID (see below)
#define REFLECTION_LAYER_GROUP												0x0000
#define REFLECTION_LAYER_MAIN_GROUP										0x0001
#define REFLECTION_LAYER_MAIN_IMAGE_GROUP							0x0002
#define REFLECTION_LAYER_MAIN_IMAGE_PREVIEW						0x0003
#define REFLECTION_LAYER_MAIN_IMAGE_MODEL							0x0004
#define REFLECTION_LAYER_MAIN_IMAGE_ANISOTROPY				0x0005
#define REFLECTION_LAYER_MAIN_IMAGE_PRIMARY						0x0006
#define REFLECTION_LAYER_MAIN_IMAGE_SECONDARY					0x0007
#define REFLECTION_LAYER_MAIN_IMAGE_DISTANCE					0x0008
#define REFLECTION_LAYER_MAIN_IMAGE_CLOTH							0x0009
//
#define REFLECTION_LAYER_MAIN_DISTRIBUTION						0x0010
#define REFLECTION_LAYER_MAIN_ADDITIVE								0x0011
#define REFLECTION_LAYER_MAIN_COLORED_REFLECTION			0x0012
#define REFLECTION_LAYER_MAIN_COLORED_SPECULAR				0x0013
#define REFLECTION_LAYER_MAIN_BLEND_MODE							0x0014
//
#define REFLECTION_LAYER_MAIN_VALUE_ROUGHNESS					0x0020
#define REFLECTION_LAYER_MAIN_VALUE_FALLOFF						0x0021
#define REFLECTION_LAYER_MAIN_VALUE_INNER							0x0022
#define REFLECTION_LAYER_MAIN_VALUE_ORENNAYAR					0x0023
#define REFLECTION_LAYER_MAIN_VALUE_REFLECTION				0x0024
#define REFLECTION_LAYER_MAIN_VALUE_SPECULAR					0x0025
//
#define REFLECTION_LAYER_MAIN_SHADER_ROUGHNESS				0x0028
#define REFLECTION_LAYER_MAIN_SHADER_FALLOFF					0x0029
#define REFLECTION_LAYER_MAIN_SHADER_INNER						0x002A
#define REFLECTION_LAYER_MAIN_SHADER_ORENNAYAR				0x002B
#define REFLECTION_LAYER_MAIN_SHADER_REFLECTION				0x002C
#define REFLECTION_LAYER_MAIN_SHADER_SPECULAR					0x002D

// Bump
#define REFLECTION_LAYER_MAIN_VALUE_BUMP							0x0030
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_MODE					0x0031
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_SPACE				0x0032
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_BUMP					0x0033
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_NORM					0x0034
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_MIPMAP				0x0035
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_REVERSEX			0x0036
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_REVERSEY			0x0037
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_REVERSEZ			0x0038
#define REFLECTION_LAYER_MAIN_VALUE_BUMP_SWAP					0x0039
//
#define REFLECTION_LAYER_MAIN_SHADER_BUMP							0x003A
#define REFLECTION_LAYER_MAIN_SHADER_BUMP_CUSTOM			0x003B

// Color
#define REFLECTION_LAYER_COLOR_GROUP									0x0040
#define REFLECTION_LAYER_COLOR_COLOR									0x0041
#define REFLECTION_LAYER_COLOR_BRIGHTNESS							0x0042
#define REFLECTION_LAYER_COLOR_TEXTURE								0x0043
#define REFLECTION_LAYER_COLOR_MIX_MODE								0x0044
#define REFLECTION_LAYER_COLOR_MIX_STRENGTH						0x0045

// Transparency
#define REFLECTION_LAYER_TRANS_GROUP									0x0050
#define REFLECTION_LAYER_TRANS_COLOR									0x0051
#define REFLECTION_LAYER_TRANS_BRIGHTNESS							0x0052
#define REFLECTION_LAYER_TRANS_TEXTURE								0x0053
#define REFLECTION_LAYER_TRANS_MIX_MODE								0x0054
#define REFLECTION_LAYER_TRANS_MIX_STRENGTH						0x0055

// Anisotropy
#define REFLECTION_LAYER_ANISO_GROUP									0x0060
#define REFLECTION_LAYER_ANISO_PROJECTION							0x0061
#define REFLECTION_LAYER_ANISO_PROJECTION_SCALE				0x0062
#define REFLECTION_LAYER_ANISO_PROJECTION_ANGLE				0x006F // id out of order
#define REFLECTION_LAYER_ANISO_PROJECTION_OFFSET_U		0x0063
#define REFLECTION_LAYER_ANISO_PROJECTION_OFFSET_V		0x0064
#define REFLECTION_LAYER_ANISO_PROJECTION_COUNT				0x0065
#define REFLECTION_LAYER_ANISO_PATTERN								0x0066
#define REFLECTION_LAYER_ANISO_MIRROR									0x0067
#define REFLECTION_LAYER_ANISO_SCRATCHES							0x0068
#define REFLECTION_LAYER_ANISO_VALUE_CUSTOM						0x0069
#define REFLECTION_LAYER_ANISO_VALUE_ANISOTROPY				0x006A
#define REFLECTION_LAYER_ANISO_VALUE_ORIENTATION			0x006B
#define REFLECTION_LAYER_ANISO_SHADER_CUSTOM					0x006C
#define REFLECTION_LAYER_ANISO_SHADER_ANISOTROPY			0x006D
#define REFLECTION_LAYER_ANISO_SHADER_ORIENTATION			0x006E
//
#define REFLECTION_LAYER_PRIMARY_VALUE_AMPLITUDE			0x0070
#define REFLECTION_LAYER_PRIMARY_VALUE_SCALE					0x0071
#define REFLECTION_LAYER_PRIMARY_VALUE_LENGTH					0x0072
#define REFLECTION_LAYER_PRIMARY_VALUE_ATTENUATION		0x0073
#define REFLECTION_LAYER_PRIMARY_SHADER_AMPLITUDE			0x0078
#define REFLECTION_LAYER_PRIMARY_SHADER_SCALE					0x0079
#define REFLECTION_LAYER_PRIMARY_SHADER_LENGTH				0x007A
#define REFLECTION_LAYER_PRIMARY_SHADER_ATTENUATION		0x007B
//
#define REFLECTION_LAYER_SECONDARY_VALUE_AMPLITUDE		0x0080
#define REFLECTION_LAYER_SECONDARY_VALUE_SCALE				0x0081
#define REFLECTION_LAYER_SECONDARY_VALUE_LENGTH				0x0082
#define REFLECTION_LAYER_SECONDARY_VALUE_ATTENUATION	0x0083
#define REFLECTION_LAYER_SECONDARY_SHADER_AMPLITUDE		0x0088
#define REFLECTION_LAYER_SECONDARY_SHADER_SCALE				0x0089
#define REFLECTION_LAYER_SECONDARY_SHADER_LENGTH			0x008A
#define REFLECTION_LAYER_SECONDARY_SHADER_ATTENUATION	0x008B

// Fresnel
#define REFLECTION_LAYER_FRESNEL_GROUP								0x0090
#define REFLECTION_LAYER_FRESNEL_MODE									0x0091
#define REFLECTION_LAYER_FRESNEL_PRESET								0x0092
#define REFLECTION_LAYER_FRESNEL_METAL								0x0093
#define REFLECTION_LAYER_FRESNEL_INVERTED							0x0094
#define REFLECTION_LAYER_FRESNEL_OPAQUE								0x009D // id out of order
#define REFLECTION_LAYER_FRESNEL_VALUE_STRENGTH				0x0095
#define REFLECTION_LAYER_FRESNEL_VALUE_IOR						0x0096
#define REFLECTION_LAYER_FRESNEL_VALUE_ETA						0x0097
#define REFLECTION_LAYER_FRESNEL_VALUE_ABSORP					0x0098
#define REFLECTION_LAYER_FRESNEL_SHADER_STRENGTH			0x0099
#define REFLECTION_LAYER_FRESNEL_SHADER_IOR						0x009A
#define REFLECTION_LAYER_FRESNEL_SHADER_ETA						0x009B
#define REFLECTION_LAYER_FRESNEL_SHADER_ABSORP				0x009C

// Sampling
#define REFLECTION_LAYER_SAMPLING_GROUP								0x00B0
#define REFLECTION_LAYER_SAMPLING_DIM									0x00B1
#define REFLECTION_LAYER_SAMPLING_DIM_DISTANCE				0x00B2
#define REFLECTION_LAYER_SAMPLING_DIM_FALLOFF					0x00B3
#define REFLECTION_LAYER_SAMPLING_CLAMP								0x00B4
#define REFLECTION_LAYER_SAMPLING_CUTOFF							0x00B5
#define REFLECTION_LAYER_SAMPLING_EXIT_COLOR					0x00B6
#define REFLECTION_LAYER_SAMPLING_SEPARATE_PASS				0x00B7
#define REFLECTION_LAYER_SAMPLING_SAMPLE_COUNT_EX			0x00B8
#define REFLECTION_LAYER_SAMPLING_DIM_COLOR						0x00B9
#define REFLECTION_LAYER_SAMPLING_SAMPLE_COUNT				0x00BA

// Cloth
#define REFLECTION_LAYER_CLOTH_GROUP									0x00C0
#define REFLECTION_LAYER_CLOTH_PRESET									0x00C1
#define REFLECTION_LAYER_CLOTH_PATTERN								0x00C2
#define REFLECTION_LAYER_CLOTH_QUALITY								0x00C3
#define REFLECTION_LAYER_CLOTH_SMOOTHING							0x00C4
#define REFLECTION_LAYER_CLOTH_SCATTERING_UNIFORM			0x00C5
#define REFLECTION_LAYER_CLOTH_SCATTERING_FORWARD			0x00C6
//
#define REFLECTION_LAYER_CLOTH_VALUE_ORIENTATION			0x00D0
#define REFLECTION_LAYER_CLOTH_VALUE_SCALE_U					0x00D1
#define REFLECTION_LAYER_CLOTH_VALUE_SCALE_V					0x00D2
#define REFLECTION_LAYER_CLOTH_VALUE_HIGHLIGHTS				0x00D3
//
#define REFLECTION_LAYER_CLOTH_SHADER_ORIENTATION			0x00D8
#define REFLECTION_LAYER_CLOTH_SHADER_SCALE_U					0x00D9
#define REFLECTION_LAYER_CLOTH_SHADER_SCALE_V					0x00DA
#define REFLECTION_LAYER_CLOTH_SHADER_HIGHLIGHTS			0x00DB
//
#define REFLECTION_LAYER_CLOTH_NOISE_INTENSITY				0x00E0
#define REFLECTION_LAYER_CLOTH_NOISE_SCALE						0x00E1
#define REFLECTION_LAYER_CLOTH_NOISE_SHAPE_WARP				0x00E2
#define REFLECTION_LAYER_CLOTH_NOISE_SHAPE_WEFT				0x00E3
#define REFLECTION_LAYER_CLOTH_NOISE_SHAPE_SCALE			0x00E4
//
#define REFLECTION_LAYER_CLOTH_COLOR_DIFFUSE_WARP			0x00E8
#define REFLECTION_LAYER_CLOTH_COLOR_DIFFUSE_WEFT			0x00E9
#define REFLECTION_LAYER_CLOTH_COLOR_SPECULAR_WARP		0x00EA
#define REFLECTION_LAYER_CLOTH_COLOR_SPECULAR_WEFT		0x00EB
//
#define REFLECTION_LAYER_CLOTH_SHADER_DIFFUSE_WARP		0x00EC
#define REFLECTION_LAYER_CLOTH_SHADER_DIFFUSE_WEFT		0x00ED
#define REFLECTION_LAYER_CLOTH_SHADER_SPECULAR_WARP		0x00EE
#define REFLECTION_LAYER_CLOTH_SHADER_SPECULAR_WEFT		0x00EF

// Debug
#define REFLECTION_LAYER_DEBUG_GROUP									0x00F0
#define REFLECTION_LAYER_DEBUG_DUDV										0x00F1
#define REFLECTION_LAYER_DEBUG_SCRATCHES							0x00F2
#define REFLECTION_LAYER_DEBUG_WARMNESS								0x00F3
#define REFLECTION_LAYER_DEBUG_CLOTH									0x00F4

class ReflectionLayer
{
public:
	String name;
	Int32	layerID;
	Int32	flags;

	// Base data ID is based on the layer id
	Int32 GetDataID() const { return REFLECTION_LAYER_LAYER_DATA + REFLECTION_LAYER_LAYER_SIZE * layerID; }
};

#endif // C4D_REFLECTION_H__
